/*
   Libre Media Server - A Media Server Sotfware for stage and performing
   Copyright (C) 2012-2013  Santiago Noreña libremediaserver@gmail.com

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef LIBREMEDIASERVER_H
#define LIBREMEDIASERVER_H

#include <QMainWindow>
#include <QApplication>
#include <QObject>
#include <QDesktopWidget>
#include <QtDebug>
#include <QtNetwork>
#include <QLocalServer>
#include <QLocalSocket>
#include <QFile>
#include <QFileInfo>
#include <QFileDialog>
#include <QTextStream>
#include <QWebView>
#include <QUrl>

#include "msex.h"
#include "CITPDefines.h"
#include "MSEXDefines.h"
#include "ui_libremediaserver.h"
#include "olainterface.h"

#define VERSION "LibreMediaServer Version 0.05-1"
#define COPYRIGHT "(C) 2012-2013 Santi Norena libremediaserver@gmail.com"
#define LICENSE "GPL 3 License. See LICENSE.txt and credits.txt for details"

class QMenu;
class QProcess;
class olaInterface;

class libreMediaServer : public QMainWindow
{

    friend class olaInterface;

    Q_OBJECT

public:

    libreMediaServer (QStringList args, QWidget *parent = 0);
    virtual ~libreMediaServer();

    Ui::LibreMediaServer ui;

protected:

    QString m_pathmedia; // Path to Medias
    olaInterface *m_olaInterface;
    QProcess *m_ola; // OLA daemon process
    QProcess *m_pd_video; // Pure Data process for video

    // Video Unix Local Sockets
    QLocalServer *m_server_vid;
    QLocalSocket *m_read_vid;
    QLocalSocket *m_write_video;

    int m_startvideo; // Counter starts video engine. Debugging purpose
    bool m_gui;

private:


    void errorsending();
    void pdstart();
    void open_start();
    void save_finish();
    void open(QFile *file);
    void save(QFile *file);
//   void MessageHandler(QtMsgType type, const char *msg);

public slots:

    void sendFrame();
    bool sendPacket(const char *buffer, int bufferLen);
    void receiveDMX(int layer, int channel, int value);
    bool newFile(QString file);

private slots:

    void olastart(); // Init the OLA daemon
    void olasetup();
    // Video
    void newPeer();
    void newmessage();
    void newconexion();
    void on_window_stateChanged(int state);
    void on_winpositionx_valueChanged();
    void on_winpositiony_valueChanged();
    void on_winsizex_valueChanged();
    void on_winsizey_valueChanged();
    void on_fpsRate_valueChanged();
    void pdrestart();
    void stdout();
    void on_previewMaster_stateChanged(int state);
    void on_previewLayer_stateChanged(int state);

    // Menu File
    void openFile();
    void saveFile();
    void ChangeMediaPath();// Change the path to medias
};

#endif // LIBREMEDIASERVER_H
