/* The CITP/MSEx 1.0 header file

   Copyright (C) 2012-2013  Santiago Noreña

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This class do all the MSEX work */

#ifndef MSEX_H
#define MSEX_H

#include <QObject>
#include <QtNetwork/QHostAddress>
#include <QtNetwork/QTcpServer>
#include <QtNetwork/QTcpSocket>
#include <QtDebug>
#include <QTimer>

#include "MSEXDefines.h"
#include "olainterface.h"
#include "citp-lib.h"
#include "libremediaserver.h"
#include "PacketCreator.h"

//class olaInterface;
class msex : public QObject
{
     Q_OBJECT
public:
    msex(olaInterface * parent);
    virtual ~msex();

    virtual QString peerName() const;
    virtual QString peerState() const;
    virtual QString peerHost() const;
    virtual quint16 peerListeningPort() const;

    QTimer *m_timer; // Frame Request

protected:

    CITPLib *m_citp;    // CITP Peer. Ploc and frame transmit
    QList<MediaLibrary> * m_media; // Link to the MediaLibrary in parent.
    unsigned char * m_buffer;
    int m_bufferLen;
    QString m_peerState;
    QString m_peerName;
    QHostAddress m_host;
    quint16 m_listeningPort;
    QTcpServer *m_tcpServer;
    QTcpSocket *m_tcpSocket;

    QList<LayerStatus> m_layers;

    bool sendPacket(const unsigned char *buffer, int bufferLen);
    bool sendPacket(const char *buffer, int bufferLen);
    bool sendNACK(quint32 header);

    // Procesado de paquetes MSEX
    void parsePacket(const QByteArray &byteArray);
    void parsePINFPacket(const QByteArray &byteArray);
    void parseMSEXPacket(const QByteArray &bytearray);
    void parseCINFPacket(const QByteArray &byteArray);
    void parseNACKPacket(const QByteArray &byteArray);
    void parseGELIPacket(const QByteArray &byteArray);
    void parseGEINPacket(const QByteArray &byteArray);
    void parseGLEIPacket(const QByteArray &byteArray);
    void parseGELTPacket();
    void parseGETHPacket(const QByteArray &byteArray);
    void parseGVSRPacket();
    void parseGELNPacket(const QByteArray &byteArray);
    void parseRQSTPacket(const QByteArray &byteArray);


private:

    QList<MediaInformation> getMediaInformation(QDir dir);

public slots:

    void sendFrame(QImage image);

protected slots:

    void handleReadyRead();

private slots:

    void newPeer();

signals:

};

#endif // MEDIASERVER_H
