////////////////////////////////////////////////////////
//   udserver v0.01
//   External for Gem and Pure Data to convert an image to jpg and send it 
//	 via Unix Domain Sockets.
//
//   (c) 2012-2013 Santi Noreña. librediaserver@gmail.com 	
//
//	 GPL License.
//
/////////////////////////////////////////////////////////
#include "m_pd.h"
#include "s_stuff.h"
#include <stdio.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <unistd.h>
#include <string.h>

#define SOCKET_PATH "/tmp/socket"
#define SOCKET_ERROR -1

static t_class *udserver_class;

typedef struct _udserver
{
    t_object x_obj;
    t_outlet *x_msgout;
    t_outlet *x_connectout;
    int x_connectsocket;
    int x_nconnections;
    int x_udp;
} t_udserver;

static void udserver_notify(t_udserver *x)
{
    outlet_float(x->x_connectout, --x->x_nconnections);
}

static void udserver_doit(void *z, t_binbuf *b)
{
    t_atom messbuf[1024];
    t_udserver *x = (t_udserver *)z;
    int msg, natom = binbuf_getnatom(b);
    t_atom *at = binbuf_getvec(b);
    for (msg = 0; msg < natom;)
    {
        int emsg;
        for (emsg = msg; emsg < natom && at[emsg].a_type != A_COMMA
            && at[emsg].a_type != A_SEMI; emsg++)
                ;
        if (emsg > msg)
        {
            int i;
            for (i = msg; i < emsg; i++)
                if (at[i].a_type == A_DOLLAR || at[i].a_type == A_DOLLSYM)
            {
                pd_error(x, "udserver: got dollar sign in message");
                goto nodice;
            }
            if (at[msg].a_type == A_FLOAT)
            {
                if (emsg > msg + 1)
                    outlet_list(x->x_msgout, 0, emsg-msg, at + msg);
                else outlet_float(x->x_msgout, at[msg].a_w.w_float);
            }
            else if (at[msg].a_type == A_SYMBOL)
                outlet_anything(x->x_msgout, at[msg].a_w.w_symbol,
                    emsg-msg-1, at + msg + 1);
        }
    nodice:
        msg = emsg + 1;
    }
}

static void udserver_connectpoll(t_udserver *x)
{
    int fd = accept(x->x_connectsocket, 0, 0);
    if (fd < 0) post("udserver: accept failed");
    else
    {
        t_socketreceiver *y = socketreceiver_new((void *)x, 
            (t_socketnotifier)udserver_notify,
                (x->x_msgout ? udserver_doit : 0), 0);
        sys_addpollfn(fd, (t_fdpollfn)socketreceiver_read, y);
        outlet_float(x->x_connectout, ++x->x_nconnections);
    }
}

static void *udserver_new(t_symbol *compatflag, t_floatarg fportno, t_floatarg udpflag)
{
    t_udserver *x;
    struct sockaddr_un server;
    int sockfd, portno;// = fportno, udp = (udpflag != 0);
    int old = !strcmp(compatflag->s_name , "old");
    int intarg;
        /* create a socket */
    sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
#if 0
    fprintf(stderr, "receive socket %d\n", sockfd);
#endif
    if (sockfd < 0)
    {
        sys_sockerror("socket");
        return (0);
    }
   // start with a clean address structure
   memset(&server, 0, sizeof(struct sockaddr_un));
   strcpy(server.sun_path, SOCKET_PATH);
   server.sun_family = AF_UNIX;
   unlink(SOCKET_PATH);   
/*
#if 1
        // ask OS to allow another Pd to repoen this port after we close it.
    intarg = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
        (char *)&intarg, sizeof(intarg)) < 0)
            post("setsockopt (SO_REUSEADDR) failed\n");
#endif
#if 0
    intarg = 0;
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF,
        &intarg, sizeof(intarg)) < 0)
            post("setsockopt (SO_RCVBUF) failed\n");
#endif
    intarg = 1;
    if(setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST, 
                  (const void *)&intarg, sizeof(intarg)) < 0)
        post("setting SO_BROADCAST");
        // Stream (TCP) sockets are set NODELAY 
    if (!udp)
    {
        intarg = 1;
        if (setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY,
            (char *)&intarg, sizeof(intarg)) < 0)
                post("setsockopt (TCP_NODELAY) failed\n");
    }
        // assign server port number 
    server.sin_port = htons((u_short)portno);
*/
        // name the socket 
    if (bind(sockfd, (struct sockaddr *)&server, sizeof(server)) < 0)
    {
        sys_sockerror("bind");
        sys_closesocket(sockfd);
        return (0);
    }
    x = (t_udserver *)pd_new(udserver_class);
    if (old)
    {
        /* old style, nonsecure version */
        x->x_msgout = 0;
    }
    else x->x_msgout = outlet_new(&x->x_obj, &s_anything);

/*    if (udp)        // datagram protocol 
    {
        t_socketreceiver *y = socketreceiver_new((void *)x, 
            (t_socketnotifier)udserver_notify,
                (x->x_msgout ? udserver_doit : 0), 1);
        sys_addpollfn(sockfd, (t_fdpollfn)socketreceiver_read, y);
        x->x_connectout = 0;
    }
    else        // streaming protocol
    {
*/        if (listen(sockfd, 5) < 0)
        {
            sys_sockerror("listen");
            sys_closesocket(sockfd);
            sockfd = -1;
        } else {
            sys_addpollfn(sockfd, (t_fdpollfn)udserver_connectpoll, x);
            x->x_connectout = outlet_new(&x->x_obj, &s_float);
        }
//    }
    x->x_connectsocket = sockfd;
    x->x_nconnections = 0;
//    x->x_udp = udp;

    return (x);
}

static void udserver_free(t_udserver *x)
{
        /* LATER make me clean up open connections */
    if (x->x_connectsocket >= 0)
    {
        sys_rmpollfn(x->x_connectsocket);
        sys_closesocket(x->x_connectsocket);
    }
}

void udserver_setup(void)
{
    udserver_class = class_new(gensym("udserver"),
        (t_newmethod)udserver_new, (t_method)udserver_free,
        sizeof(t_udserver), CLASS_NOINLET, A_DEFFLOAT, A_DEFFLOAT, 
            A_DEFSYM, 0);
}


